/*  MARKOVM.PRG

This is a GAUSS program.  It will implement the estimation and testing
procedures for the Markov switching mean AR(p) model of Hamilton (1989)
as presented in B. Hansen "The likelihood ratio test under non-standard
conditions: Testing the Markov trend model of GNP."

Written November, 1991
Modified May,  1994

written by:

Bruce E. Hansen
Department of Economics
Boston College
Chestnut Hill, MA 02167-3806

(617) 552-3678
Hansen-EC@hermes.bc.edu



The following parameters need to be specified:

dat = data
nar = 4            sets number of autoregressive lags

gn = number of gridpoints

_tnull = 1  to test the null of one state
_tnull = 0  to not conduct this test
_nwband = 5   (e.g.) Sets the (Newey-West) maximal covariance bandwidth to 5

_gopt  = 1  to calculate global optimum
_gopt  = 0  to not calculate global optimum


Note:  If _gopt = 1  and _tnull = 0, then one needs to specify starting values
       for the parameters in  "th".  The order is

(1) Mean in state 0
(2) AR coefficients
(3) Standard deviation of error in state 0
(4) Difference in mean between states
(5) p
(6) q


If _tnull = 1, then it is necessary to specify a grid for the differences in
the mean between states, for p, and for q.
These gridpoints go in the matrices.

  "gx", "gp" and "gq"

where gp and gq are vectors, and gx is gn x 1 matrix, gn being number of
gridpoints.

Finally, an appropriate likelihood function must be included.  This likelihood
function be a function of all of the above listed variables, even when not
relevant.  The function should return a vector which yields the likelihood
evaluated at each observation, i.e., an nx1 vector.

This program is currently set to reproduce the results from Hansen's paper
using his "Grid 3".                                                       */


/* ***********************************/
output file = "markovm.out" reset;
output off;
n = 135;
load gnp[]=gnp82.dat;
gnp = 100*ln(gnp);
dat = gnp[2:n+1] - gnp[1:n];

nar = 4;
_tnull = 1;
_gopt = 0;
_nwband = 5;

let th = -.4 .01 -.06 -.25 -.21 .77 1.5 2.2 1.1 ;
@let th = -.4 .77 1.5 2.2 1.1 ;@

gn = 20;
gx = seqa(.1,.1,gn);

library optmum;
optset;
#include optmum.ext;
_opmiter = 500;
_opgtol = .0001;

pn = 8;
gp = seqa(.12,.11,pn);
qn = pn;
gq = gp;

/* *****************************************/

b1 = 0;
p = 0.5;
q = 0.5;

y = dat[1+nar:n];
if nar .> 0;
  x = dat[nar:n-1];
  it = 2; do while it <= nar;
    x = x~dat[nar-it+1:n-it];
  it = it + 1; endo;
  z1 = 0|1;
  n3 = 2.^(nar-1);
  ss = z1.*.ones(n3,1);
  it = 2; do while it <= nar;
    ss = ss~(ones(2.^(it-1),1).*.(z1.*.ones(2.^(nar-it),1)));
  it = it + 1; endo;
  ss0 = ss[1:n3,.];
  ss1 = ss[n3+1:n3*2,.];
endif;

if cols(gx) .ne 1;
"";"";
"Columns of matrix  gx  must equal 1";
"";
"Program Terminates";
end;
endif;

proc marklike(th);
local c,alpha0,phi,sig0,alpha1,sig1,p,q,rho,b8,pxa,pxb,pa,pb,pp,z0,z1,p0,p1,
      q00,q01,q10,q11,f00,f01,f10,f11,fit,it,ff,f,nr,s1,sphi0,sphi1,z;

c = sqrt(2*pi);
alpha0 = th[1];
sig0 = th[2+nar];
alpha1 = th[3+nar];
p = th[4+nar];
q = th[5+nar];
rho = (1-q)/(2-p-q);

if nar .> 0;
  phi = th[2:1+nar];
  b8 = alpha0*(1 - sumc(phi));
  sphi0 = ss0*phi*alpha1;
  sphi1 = ss1*phi*alpha1;
else;
  b8 = alpha0;
endif;

pp = (1-rho)|rho;
pxa = q|(1-p);
pxb = 1-pxa;

it = 2; do while it <= nar;
  nr = ones((2.^(it-2)),1);
  pp = (pp.*(pxa.*.nr))|(pp.*(pxb.*.nr));
it = it + 1; endo;
nr = rows(pp)/2;
p0 = pp[1:nr];
p1 = pp[nr+1:2*nr];

fit = zeros(n-nar,1);
it = 1;
do until it > n-nar;

  z = y[it] - b8;
  if nar .> 0;
    z = z - x[it,.]*phi;
    z0 = z + sphi0;
    z1 = z + sphi1;
  else;
    z0 = z;
    z1 = z;
  endif;

  f00 = exp(- (z0.^2)/(2*sig0*sig0))./(sig0*c);
  f01 = exp(- (z1.^2)/(2*sig0*sig0))./(sig0*c);
  f10 = exp(- ((z0-alpha1).^2)/(2*sig0*sig0))./(sig0*c);
  f11 = exp(- ((z1-alpha1).^2)/(2*sig0*sig0))./(sig0*c);

  q00 = (f00.*p0)*q;
  q01 = (f01.*p1)*(1-p);

  q10 = (f10.*p0)*(1-q);
  q11 = (f11.*p1)*p;

  if nar .> 1;
    s1 = seqa(1,2,rows(p0)/2);
    p0 = (q00[s1] + q00[s1+1])|(q01[s1] + q01[s1+1]);
    p1 = (q10[s1] + q10[s1+1])|(q11[s1] + q11[s1+1]);
  else;
    p0 = q00 + q01;
    p1 = q10 + q11;
  endif;
  ff = sumc(p0) + sumc(p1);
  p0 = p0./ff;
  p1 = p1./ff;
  fit[it] = ff;

it = it+1;
endo;
if minc(fit) .> 0;
 f = ln(fit);
else;
 f = -1000;
endif;

retp(-f);
endp;

proc dmclike(th);
local c,alpha0,phi,sig0,alpha1,sig1,rho,b8,pxa,pxb,pa,pb,pp,z0,z1,p0,p1,
      q00,q01,q10,q11,f00,f01,f10,f11,fit,it,ff,f,nr,s1,sphi0,sphi1,z,
      grdf,phisc,sig2,xit,dp0,dp1,gradf,df,xit0,xit1,dlnf,
      df00,df01,df10,df11,dq00,dq01,dq10,dq11,z10,z11,z02,z12,z102,z112,
   dfa00,dfa01,dfa10,dfa11,dfp00,dfp01,dfp10,dfp11,dfs00,dfs01,dfs10,dfs11;

c = sqrt(2*pi);
alpha0 = th[1];
sig0 = th[2+nar];
sig2 = sig0.^2;
alpha1 = b1;
rho = (1-q)/(2-p-q);

phisc = 1;
if nar .> 0;
  phi   = th[2:1+nar];
  sphi0 = ss0*phi*alpha1;
  sphi1 = ss1*phi*alpha1;
  phisc = 1 - sumc(phi);
endif;
b8 = alpha0*phisc;

pp = (1-rho)|rho;
pxa = q|(1-p);
pxb = 1-pxa;

it = 2; do while it <= nar;
  nr = ones((2.^(it-2)),1);
  pp = (pp.*(pxa.*.nr))|(pp.*(pxb.*.nr));
it = it + 1; endo;
nr = rows(pp)/2;
p0 = pp[1:nr];
p1 = pp[nr+1:2*nr];
dp0 = zeros(nr,2+nar);
dp1 = zeros(nr,2+nar);

grdf = zeros(n-nar,2+nar);
it = 1;  do until it > n-nar;

  z = y[it] - b8;
  if nar .> 0;
    xit = x[it,.];
    z = z - xit*phi;
    z0 = z + sphi0;
    z1 = z + sphi1;
  else;
    z0 = z;
    z1 = z;
  endif;
    z10 = z0 - alpha1;
    z11 = z1 - alpha1;
    z02 = z0.^2;
    z12 = z1.^2;
    z102 = z10.^2;
    z112 = z11.^2;

  f00 = exp(- z02/(2*sig2))./(sig0*c);
  f01 = exp(- z12/(2*sig2))./(sig0*c);
  f10 = exp(- z102/(2*sig2))./(sig0*c);
  f11 = exp(- z112/(2*sig2))./(sig0*c);

@  These should be the derivatives of ln f(y|S)  @

@ With respect to alpha0 : @

  dfa00 =  z0.*phisc./sig2;
  dfa01 =  z1.*phisc./sig2;
  dfa10 = z10.*phisc./sig2;
  dfa11 = z11.*phisc./sig2;

@ With resepct to phi : @

  if nar .> 0;
    xit0 = xit - ss0.*alpha1 - alpha0;
    xit1 = xit - ss1.*alpha1 - alpha0;
    dfp00 = z0.*xit0./sig2;
    dfp01 = z1.*xit1./sig2;
    dfp10 = z10.*xit0./sig2;
    dfp11 = z11.*xit1./sig2;
  endif;

@ With resepct to sig0 : @

  dfs00 = z02./(sig0*sig2) - 1/sig0;
  dfs01 = z12./(sig0*sig2) - 1/sig0;
  dfs10 = z102./(sig0*sig2) - 1/sig0;
  dfs11 = z112./(sig0*sig2) - 1/sig0;

  if nar .> 0;
    df00 = dfa00~dfp00~dfs00;
    df01 = dfa01~dfp01~dfs01;
    df10 = dfa10~dfp10~dfs10;
    df11 = dfa11~dfp11~dfs11;
  else;
    df00 = dfa00~dfs00;
    df01 = dfa01~dfs01;
    df10 = dfa10~dfs10;
    df11 = dfa11~dfs11;
  endif;

  q00 = (f00.*p0)*q;
  q01 = (f01.*p1)*(1-p);
  q10 = (f10.*p0)*(1-q);
  q11 = (f11.*p1)*p;

  if nar .> 1;
    s1 = seqa(1,2,rows(p0)/2);
    p0 = (q00[s1] + q00[s1+1])|(q01[s1] + q01[s1+1]);
    p1 = (q10[s1] + q10[s1+1])|(q11[s1] + q11[s1+1]);
  else;
    p0 = q00 + q01;
    p1 = q10 + q11;
  endif;
  ff = sumc(p0) + sumc(p1);
  p0 = p0./ff;
  p1 = p1./ff;

  dq00 = q00.*(dp0 + df00);
  dq01 = q01.*(dp1 + df01);
  dq10 = q10.*(dp0 + df10);
  dq11 = q11.*(dp1 + df11);

  df = sumc(dq00) + sumc(dq01) + sumc(dq10) + sumc(dq11);
  dlnf = df/ff;
  grdf[it,.] = dlnf';

if nar .> 1;
  s1 = seqa(1,2,rows(p0)/2);
  dp0 = ((dq00[s1,.] + dq00[s1+1,.])|(dq01[s1,.] + dq01[s1+1,.]));
  dp1 = ((dq10[s1,.] + dq10[s1+1,.])|(dq11[s1,.] + dq11[s1+1,.]));
else;
  dp0 = dq00 + dq01;
  dp1 = dq10 + dq11;
endif;
dp0 = dp0./(p0.*ff) - dlnf';
dp1 = dp1./(p1.*ff) - dlnf';


it = it + 1;
endo;

retp(-sumc(grdf)');
endp;

@  Calculation Under null  @

if nar .> 0;
  xx = ones(n-nar,1)~x;
else;
  xx = ones(n-nar,1);
endif;
xxi = invpd(moment(xx,0));
b0 = xxi*(xx'y);
u = y - xx*b0;
v0 = sqrt((u'u)/rows(u));
var0 = xxi*((xx.*u)'(xx.*u))*xxi;
se0 = sqrt(diag(var0));
if nar .> 0;
 phi = b0[2:1+nar];
 a0 = b0[1]/(1-sumc(phi));
 b = a0|phi|v0;
else;
 a0 = b0[1];
  b = a0|v0;
endif;

proc clike(th);
local ths;
  ths = th|b1|p|q;
retp(marklike(ths));
endp;

proc mclike(th);
local lik;
lik = sumc(clike(th));
retp(lik);
endp;

proc glike(th);
local lik,p,q,th2;
p = th[4+nar];
q = th[5+nar];
p = exp(p)/(1+exp(p));
q = exp(q)/(1+exp(q));
th2 = th[1:3+nar]|p|q;
lik = marklike(th2);
retp(lik);
endp;

proc mglike(th);
local lik;
lik = sumc(glike(th));
retp(lik);
endp;

b1 = 0;
null = clike(b);
"Null" sumc(null);

if _tnull == 1;

t1 = hsec;
__output = 0;
format 12,6;

gnn = gn;
crit = zeros(rows(y),pn*qn*gnn);
if _gopt == 1;
 beta = zeros(5+nar,pn*qn*gnn);
endif;
j = 0;
_opgdprc = &dmclike;

i1 = 1; do while i1 <= rows(gp);
p = gp[i1];

i2 = 1; do while i2 <= rows(gq);
q = gq[i2];

bs = b;
bi = 1;

do while maxc(bi) <= gn;
  b1 = diag(gx[bi,.]);
   j = j + 1;
   {bnew,f,gr,retcode} = optmum(&mclike,bs);
   tm = _opitdta[2]*60;
   p~q~b1~_opitdta[1]~tm;
  if _gopt == 1;
   beta[1:2+nar,j] = bnew;
   beta[3+nar:3+nar,j] = b1;
   beta[4+nar,j] = p;
   beta[5+nar,j] = q;
  endif;
   crit[.,j] = null - clike(bnew);
  bs = bnew;
  bi[1] = bi[1] + 1;
  if bi[1] > gn;
   bs = b;
   if minc(bi) < gn;
    do until maxc(bi) <= gn;
     ji = maxindc(bi);
     bi[ji] = 1;
     bi[ji-1] = bi[ji-1] + 1;
    endo;
   endif;
  endif;
endo;

i2 = i2 + 1; endo;
i1 = i1 + 1; endo;

c = sumc(crit);
nr = rows(crit);
crit = crit - (c')/nr;
se = sqrt(sumc(crit.^2));
crit = crit./(se');
cse = c./se;
mc = maxindc(cse);
m = cse[mc];

rep = 1000;
u = zeros(rep,_nwband+1);
i1 = 1; do while i1 <= rep;
   e = rndn(nr+_nwband,1);
   ee = e[1:nr]*ones(1,_nwband+1);
   i2 = 1; do while i2 <= _nwband;
     ee[.,i2+1:_nwband+1] = ee[.,i2+1:_nwband+1]
       +  e[1+i2:nr+i2]*ones(1,_nwband-i2+1);
   i2 = i2 + 1; endo;
   ee = ee./sqrt(seqa(1,1,_nwband+1)');
   u[i1,.] = maxc(crit'ee)';
i1 = i1 + 1; endo;

pvalue = meanc(u .> m);
i = 1; do while i<= _nwband+1;
  u[.,i] = sortc(u[.,i],1);
i = i + 1; endo;
let cr = .8 .9 .95 .99;
cu = u[round(cr*rep),.];

t2 = hsec;

endif;

if _gopt == 1;

t3 = hsec;

if _tnull == 1;
 th = beta[.,maxindc(c)];
 p = th[4+nar];
 th[4+nar] = ln(p/(1-p));
 q = th[5+nar];
 th[5+nar] = ln(q/(1-q));
endif;

__output = 2;
{bnew,f,gr,retcode} = optmum(&mglike,th);
lr = sumc(null) - f;

 g = gradp(&glike,bnew);
 s = g'g;
 h = pinv(hessp(&mglike,bnew)) ;
 v = h*s*h;

 p = bnew[4+nar];
 q = bnew[5+nar];
 p = exp(p)/(1+exp(p));
 q = exp(q)/(1+exp(q));
 bopt = bnew[1:3+nar]|p|q;
 vt = (p*(1-p))|(q*(1-q));
 v[4+nar 5+nar,3+nar 4+nar] = vt.*v[4+nar 5+nar,3+nar 5+nar].*(vt');

 se = sqrt(diag(v));
 g = cumsumc(g);
 stable = sumc(diag((g'g)/s))/rows(g);

t4 = hsec;
endif;

output on;
"Estimation Under Null of Linearity";
" Regression Parameters:"
b0~se0;
"";
"Standard Deviation : " v0;
"";
"Log Likelihood " sumc(null);
"";
if _gopt == 1;
"Results from estimation and testing of Markov switching model";
"";
"Estimates and standard errors";
"";
"Regression Parameters in State 0";
bopt[1:1+nar]~se[1:1+nar];
"Standard deviation";
bopt[2+nar]~se[2+nar];
"";
"Difference in mean between states";
bopt[3+nar]~se[3+nar];
"P";
bopt[4+nar]~se[4+nar];
"Q";
bopt[5+nar]~se[5+nar];
"";"";
"Likelihood Value " f;
"Likelihood Ratio " lr;
"2*LR             " 2*lr;
"Nyblom Stability " stable;
"Estimation Time (Hours) " (t4-t3)/360000;
"";"";
endif;
if _tnull == 1;
"Test of Null of One-State Model";
"";
"Grid for p";
pn';
"";
"Grid for Q";
qn';
"";
"Grid for mean";
gx;
"";
"Internal Monte Carlo Replications" rep;
"";
"Standardized LR Test   " m;
" P-values              ";
seqa(0,1,_nwband+1)~pvalue;
" Upper Criticals";
((1-cr)')|(cu');
if _gopt == 1;
 "Standardized LR Maximized at";
 th = beta[.,mc];
 "  Mean,                    " (th[3+nar]');
 "  P, Q                     " (th[4+nar:5+nar]');
 "";
endif;
"Computation Time (Hours) " (t2-t1)/360000;
"";"";"";"";
endif;
output off;
